// DisasmFactory.cpp: implementation of the CDisasmFactory class.
//
//////////////////////////////////////////////////////////////////////

#define STRICT

#include "stdafx.h"
#include "DisasmFactory.h"

#include "libdasm.h"	// include libdasm.h 
#include "distorm.h"
#include "disasm.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#pragma hdrstop

// Link the library into our project.

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDisasmFactory::CDisasmFactory()
{

}

CDisasmFactory::~CDisasmFactory()
{

}


/*
FUNCTION: FlushDecoded
*/
void CDisasmFactory::FlushDecoded(DISASSEMBLY *Disasm)
{
	// Clear all information of an decoded 
	// Instruction
	
    strcpy(Disasm->Assembly,""); // Clear menemonic
    strcpy(Disasm->Remarks,"");  // Clear commets
    strcpy(Disasm->Opcode,"");   // Clear opcodes linear
    Disasm->OpcodeSize=1;        // Smallest opcode size
	Disasm->PrefixSize=0;        // No Prefixes
}

int CDisasmFactory::ConvertLibdasm(CString &strOpcode, CString &strDisasm, BOOL bCodeOnly, DWORD nInitialAddress)
{
	DISASSEMBLY DisasmData;
	DWORD Index=0;
	
	// ؽƮ ͼ Ʈ ڵ ȯѴ.
	char *Opcodes = NULL;
	int nSize = 0;
	Opcodes = (char*) malloc(sizeof(char) * (strOpcode.GetLength()));
	nSize = OpcodeToBinary(strOpcode, (BYTE*)Opcodes);
	
	if (nSize == 0)
	{
		return 0;
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Load Library
	//////////////////////////////////////////////////////////////////////////
	
	HMODULE hDll;
	hDll = LoadLibrary("DisasmEngineDLL.dll"); // Load the DLL.
	if(hDll == NULL)
		return -1;
	
	// Get Fucntion Pointer
	DISASM_FUNCTION Decode = (DISASM_FUNCTION)GetProcAddress((HINSTANCE)hDll,"Disassemble");
	
	if(Decode==NULL) // no pointer
	{
		FreeLibrary(hDll);
		return -1;
	}
	
	DisasmData.Address=nInitialAddress; // intial address
	FlushDecoded(&DisasmData);     // clear the struct
	
	CString strSubDecoded;
	strDisasm.Empty();
	
	for(Index=0;Index<nSize;Index++)
	{
		// Decode Instruction(s)
		Decode(&DisasmData,Opcodes,&Index);
		
		// Upper case Assembly (Optional)
		CharUpper(DisasmData.Assembly);
		// Show Decoded instruction, size, remarks...
		if (!bCodeOnly)
		{
			strSubDecoded.Format("%08X %-22s\t%-20s",DisasmData.Address, DisasmData.Opcode, DisasmData.Assembly);
			if (DisasmData.Remarks[0] != '\0')
			{
				strSubDecoded += ";";
				strSubDecoded += DisasmData.Remarks;
			}
		}
		else
		{
			strSubDecoded.Format("%-20s",DisasmData.Assembly);
		}

		strDisasm += strSubDecoded;
		strDisasm += "\r\n";
		// Calculate total Size of an instruction + Prefixes, and
		// Fix the address of eIP 
		DisasmData.Address+=DisasmData.OpcodeSize+DisasmData.PrefixSize;
		// Clear all information
		FlushDecoded(&DisasmData);
	}
	
	// free libray in the end
	// of the process
	if(hDll!=NULL)
	{
		FreeLibrary(hDll);	
	}

	free(Opcodes);

	return nSize;
}

int CDisasmFactory::Disasm(CString &strOpcode, CString &strDisasm, int nMode, BOOL bCodeOnly, DWORD nInitialAddress)
{
	switch(nMode)
	{
	case DISASSEMBLE_LIBDASM:
		return ConvertLibdasm(strOpcode, strDisasm, bCodeOnly, nInitialAddress);
		break;
	case DISASSEMBLE_DISTDASM:
		return ConvertDistasmEngine(strOpcode, strDisasm, bCodeOnly, nInitialAddress);
		break;
	case DISASSEMBLE_DISASMENGINE:
		return ConvertDisasmEngine(strOpcode, strDisasm, bCodeOnly, nInitialAddress);
		break;
	}

	return -1;
}

int CDisasmFactory::ConvertDisasmEngine(CString &strOpcode, CString &strDisasm, BOOL bCodeOnly, DWORD nInitialAddress)
{
	DWORD Index=0;

	// ؽƮ ͼ Ʈ ڵ ȯѴ.
	BYTE *Opcodes = NULL;
	int nSize = 0;
	Opcodes = (BYTE*) malloc(sizeof(BYTE) * (strOpcode.GetLength()));
	nSize = OpcodeToBinary(strOpcode, Opcodes);

	if (nSize == 0)
	{
		return 0;
	}

	INSTRUCTION inst;	// declare struct INSTRUCTION
	int i, c = 0, bytes, len;
	Format format = FORMAT_INTEL;
	char string[256];
	bytes = 11;

	strDisasm.Empty();
	CString strTemp;

	while (c < nSize) {
		/*
		 * get_instruction() has the following parameters:
		 *
		 * - &inst is a pointer to struct INSTRUCTION
		 * - Opcodes + c is pointer to data to be disassembled
		 * - disassemble in 32-bit mode: MODE_32 
		 */
		len = get_instruction(&inst, Opcodes + c, MODE_32);
		
		// Illegal opcode or opcode longer than remaining buffer
		if (!len || (len + c > nSize)) {
			if(!bCodeOnly) 
			{
				strTemp.Format("%.8x ", c);
				strDisasm += strTemp;

				if (bytes) {
					strTemp.Format("%.2x  ", Opcodes[c]);
					strDisasm += strTemp;

					for (i = 1; i < bytes*2 - 1; i++)
						strDisasm += " ";
				}
				if (format == FORMAT_INTEL)
				{
					strTemp.Format("db 0x%.2x\r\n", *(Opcodes + c));
					strDisasm += strTemp;
				}
				else
				{
					strTemp.Format(".byte 0x%.2x\r\n", *(Opcodes + c));
					strDisasm += strTemp;
				}
			}
			c++;
			continue;
		}

		/*
		* Print absolute offset and raw Opcodes bytes up to 'bytes'
		* (not needed, but looks nice).
		*
		*/

		if (!bCodeOnly)
		{
			strTemp.Format("%.8x ", c);
			strDisasm += strTemp;
			if (bytes) {
				for (i = 0; i < MIN(bytes, len); i++)
				{
					strTemp.Format("%.2x", Opcodes[c + i]);
					strDisasm += strTemp;
				}
				strDisasm += " ";
				for (i = MIN(bytes, len); i < bytes*2 - len; i++)
					strDisasm += " ";
			}
		}

		/*
		* Print the parsed instruction, format using user-supplied
		* format. We could of course format the instruction in some
		* other way by accessing struct INSTRUCTION members directly.
		*/
		get_instruction_string(&inst, format, (DWORD)c, string, sizeof(string));
		strTemp.Format("%s\r\n", string);
		strDisasm += strTemp;
		c += len;
	} 
    
	free(Opcodes);

	return nSize;
}

int CDisasmFactory::ConvertDistasmEngine(CString &strOpcode, CString &strDisasm, BOOL bCodeOnly, DWORD nInitialAddress)
{
	BYTE *Opcodes = NULL;
	BYTE *Opcodes_original = NULL;
	int nSize = 0;

	Opcodes_original = Opcodes = (BYTE*) malloc(sizeof(BYTE) * (strOpcode.GetLength()));
	nSize = OpcodeToBinary(strOpcode, Opcodes);

	if (nSize == 0)
	{
		return 0;
	}

	// Version of used compiled library.
	unsigned long dver = 0;
	// Holds the result of the decoding.
	_DecodeResult res;
	// Decoded instruction information.
	_DecodedInst decodedInstructions[MAX_INSTRUCTIONS];
	// next is used for instruction's offset synchronization.
	// decodedInstructionsCount holds the count of filled instructions' array by the decoder.
	unsigned int decodedInstructionsCount = 0, i, next;

	// Default decoding mode is 32 bits, could be set by command line.
	_DecodeType dt = Decode32Bits;

	// Default offset for buffer is 0, could be set in command line.
	_OffsetType offset = 0;
	char* errch = NULL;

	CString strTemp;
	strDisasm.Empty();

	// Decode the buffer at given offset (virtual address).
	while (1) {
		// If you get an unresolved external symbol linker error for the following line,
		// change the SUPPORT_64BIT_OFFSET in distorm.h.
		res = distorm_decode(offset, (const unsigned char*)Opcodes, nSize, dt, decodedInstructions, MAX_INSTRUCTIONS, &decodedInstructionsCount);
		if (res == DECRES_INPUTERR) {
			// Null buffer? Decode type not 16/32/64?
			strDisasm = "Input error, halting!";
		}

		for (i = 0; i < decodedInstructionsCount; i++) {
#ifdef SUPPORT_64BIT_OFFSET
			if (!bCodeOnly)
				strTemp.Format("%0*I64x %-22s\t%s%s%s\r\n", dt != Decode64Bits ? 8 : 16, decodedInstructions[i].offset, (char*)decodedInstructions[i].instructionHex.p, (char*)decodedInstructions[i].mnemonic.p, decodedInstructions[i].operands.length != 0 ? " " : "", (char*)decodedInstructions[i].operands.p);
			else
				strTemp.Format("%s%s%s\r\n", (char*)decodedInstructions[i].mnemonic.p, decodedInstructions[i].operands.length != 0 ? " " : "", (char*)decodedInstructions[i].operands.p);
			strDisasm += strTemp;
#else
			if (!bCodeOnly)
				strTemp.Format("%08x %-15s\t%s%s%s\r\n", decodedInstructions[i].offset, (char*)decodedInstructions[i].instructionHex.p, (char*)decodedInstructions[i].mnemonic.p, decodedInstructions[i].operands.length != 0 ? " " : "", (char*)decodedInstructions[i].operands.p);
			else
				strTemp.Format("%s%s%s\r\n", (char*)decodedInstructions[i].mnemonic.p, decodedInstructions[i].operands.length != 0 ? " " : "", (char*)decodedInstructions[i].operands.p);
			strDisasm += strTemp;
#endif
		}

		if (res == DECRES_SUCCESS) break; // All instructions were decoded.
		else if (decodedInstructionsCount == 0) break;

		// Synchronize:
		next = (unsigned long)(decodedInstructions[decodedInstructionsCount-1].offset - offset);
		next += decodedInstructions[decodedInstructionsCount-1].size;
		// Advance ptr and recalc offset.
		Opcodes += next;
		nSize -= next;
		offset += next;
	}

	// Release buffer
	free(Opcodes_original);

	return nSize;
}


CString CDisasmFactory::BinaryToOpcode(BYTE *byteBinary, int nSize, int nLineLength, int nMode, BOOL bMakeAssembleUpper)
{
	CString strRet;
	
	if (nSize <= 0)
		return strRet;
	
	// Get a byte and print it out as a hex value
	//
	int count = 1;
	char temp[5] = {0,};
	
	int index = 0;
	for (index = 0; index < nSize; index++)
	{
		if (index % nLineLength == 0)
		{
			if(nMode == ASSEMBLE_SHELLCODE) strRet += "\"";
		}

		if (nMode == ASSEMBLE_SHELLCODE) wsprintf(temp, "\\x%02x", byteBinary[index]);
		else if (nMode == ASSEMBLE_HEX_SPACE) wsprintf(temp, "%02x ", byteBinary[index]);
		else if (nMode == ASSEMBLE_HEX_NOSPACE) wsprintf(temp, "%02x", byteBinary[index]);

		if (bMakeAssembleUpper)
		{
			int nLength = strlen(temp);
			for (int pos = 0 ; pos < nLength; pos++)
			{
				if (temp[pos] >= 'a' && temp[pos] <= 'f')
					temp[pos] = 'A' + (temp[pos] - 'a');
			}
		}

		strRet += temp;
		if ((index+1) % nLineLength == 0)
		{
			if (nMode == ASSEMBLE_SHELLCODE) strRet += "\"\r\n";
			else if (nMode == ASSEMBLE_HEX_SPACE) strRet += "\r\n";
			else if (nMode == ASSEMBLE_HEX_NOSPACE) strRet += "\r\n";
		}
	}

	if (index % nLineLength != 0)
	{
		if (nMode == ASSEMBLE_SHELLCODE) strRet += "\"";
	}

	return strRet;
}

int CDisasmFactory::OpcodeToBinary(CString strOpcode, BYTE *byteBinary)
{
	int nSize = 0;
	int nPos = 0;
	int nLength = strOpcode.GetLength();
	strOpcode.MakeUpper();

	CString strHexValue;
	
	int opcode_type = OPCODE_GENERIC;
	if (strOpcode.Find("\\X", 0) != -1)
	{
		opcode_type = OPCODE_SHELLCODE;
	}
	else if (strOpcode.Find("<BH:", 0) != -1)
	{
		opcode_type = OPCODE_FROM_FRHED;
	}

	while(nPos < nLength)
	{
		switch(opcode_type)
		{
		case OPCODE_GENERIC:
			{
				if ((strOpcode[nPos] >= '0' && strOpcode[nPos] <= '9') ||
					(strOpcode[nPos] >= 'A' && strOpcode[nPos] <= 'F'))
				{
					if (nPos+1 >= nLength)
						break;

					strHexValue.Format("%c%c",strOpcode[nPos], strOpcode[nPos+1]);			

					byteBinary[nSize] = (BYTE) HexToLong(strHexValue);
					nSize++;
					strHexValue.Empty();
					nPos += 2;
				}
				else
				{
					nPos++;
				}
			}
			break;
		case OPCODE_SHELLCODE:
			{
				if (strOpcode[nPos] == '\\' && strOpcode[nPos+1] == 'X')
				{
					if (nPos+3 >= nLength)
						break;

					strHexValue.Format("%c%c",strOpcode[nPos+2], strOpcode[nPos+3]);			

					byteBinary[nSize] = (BYTE) HexToLong(strHexValue);
					nSize++;
					strHexValue.Empty();
					nPos += 4;
				}
				else
				{
					nPos++;
				}
			}
			break;
		case OPCODE_FROM_FRHED:
			{
				if (strOpcode[nPos] == '<' && strOpcode[nPos+1] == 'B' && strOpcode[nPos+2] == 'H' && strOpcode[nPos+3] == ':')
				{
					if (nPos+6 >= nLength)
						break;

					strHexValue.Format("%c%c",strOpcode[nPos+4], strOpcode[nPos+5]);			

					byteBinary[nSize] = (BYTE) HexToLong(strHexValue);
					nSize++;
					strHexValue.Empty();
					nPos += 7;
				}
				else
				{
					nPos++;
				}
			}
		}
	}
	
	return nSize;
}

long CDisasmFactory::HexToLong(CString &szHexValue)
{
	szHexValue.MakeUpper();
	szHexValue.TrimLeft();
	szHexValue.TrimRight();
	
	long lDigitCount = szHexValue.GetLength();
	long lValue = 0;
	for (long lDigit = 0; lDigit < lDigitCount; ++lDigit)
	{
		lValue *= 16;
		
		TCHAR cDigit = szHexValue.GetAt(lDigit);
		if (cDigit >= '0' && cDigit <= '9')
			lValue += cDigit - '0';
		else if (cDigit >= 'A' && cDigit <= 'F')
			lValue += 10 + cDigit - 'A';
		else
			throw "Bad digit in Hex value string\nHexToLong()";
	}
	
	return lValue;
}

int CDisasmFactory::Assem(CStringArray &arrAsm, BYTE *byteBinary)
{
	// INC ۵  -_-;
	// a.exe 

	t_asmmodel am;
	char errtext[TEXTLEN];

	int nLine = arrAsm.GetSize();
	int nSize = 0;
	CString strLine;

	for (int line = 0 ; line < nLine ; line++)
	{
		strLine = arrAsm.GetAt(line);
		strLine.Trim();
		strLine.MakeUpper();

		int j=Assemble((LPSTR)(LPCSTR)strLine,0x400000,&am,0,0,errtext);
		for (int i=0; i<j; i++) 
		{
			byteBinary[nSize] = (BYTE)am.code[i];
			nSize++;
		}
	}

	return nSize;
}